/****************************************************************************
** pm.c
****************************************************************************
*/ 

#define EXTERN extern

/*------------------------
** Include files 
**------------------------
*/
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sys/ioctl.h> 
#include <sys/time.h> 
   
#include <pmac.h>   
#include <pm.h>   

#include "ir1.h"

static double pm_elapse_sec ( struct timeval *start, struct timeval *end);

/*----------------------------------------------------------------------
**  pm_apos_set() - Initialize the absolute position in the pmac
**    by issuing the '$*' command. Leave the pmac in OpenLoop.
**----------------------------------------------------------------------
*/
int pm_apos_set( int fd, 
   long ra, long dec        // ra,dec - in counts.
) 
{
   int rc;
   char cmd[124];

   sprintf( cmd, " M3100=$%lx M3101=$%lx M3200=$%lx M3201=$%lx #1 $* #2 $* ", 
            (ra & 0xffffff),  (ra>>24 & 0xff),
            (dec & 0xffffff), (dec>>24 & 0xff));

   printf("ra  %ld  0x%lx \n", ra, ra);
   printf("dec %ld  0x%lx \n", dec, dec);
   printf("cmd [%s] \n", cmd);
   rc = pmac_reg_cmd( fd, cmd, cmd, sizeof(cmd));
   return (rc ? ERR_DEVICE_IO : ERR_NONE);
}

/*----------------------------------------------------------------------
**  pm_init() - Initialize the some PMAC operations for T3 
**----------------------------------------------------------------------
*/
int pm_init( int fd )
{
   int rc;

   /* initialize the variable buffer setup */
   rc = pm_vb_init( fd );
   if( rc )
      return rc;

   return ERR_NONE;
}

/***********************************************************************/
/*                  mp_vb - Variable Buffer init function              */
/***********************************************************************/

/*----------------------------------------------------------------------
**  pm_vb_init() - Initialize the DPR Variable Buffer setup for T3
**----------------------------------------------------------------------
*/
int pm_vb_init( int fd )
{
   int rc;
   struct pmac_vb_cntl_t vbc;
   char   buf[80];

   /* VB Read Buffer Setup array */
   static long vb_read_setup[PM_VB_READ_BUF_NITEMS] = {
      // -----------------------  Motor1:
      PMAC_VB_Y | 0x000089,    // position scale factor (Ixx08)
      PMAC_VB_Y | 0x00009c,    // vel loop scale factor (Ixx09)
      PMAC_VB_D | 0x000088,    // Motor1 present desired position
      PMAC_VB_X | 0x000092,    // Motor1 PID D term (Ixx31)
      PMAC_VB_Y | 0x000099,    // Motor1 PID P term (Ixx30)
      PMAC_VB_X | 0x00009f,    // Motor1 PID I term (Ixx33)
      PMAC_VB_Y | 0x078006,    // ADC1 - 16 bit ADC input (DC Tach value)
      PMAC_VB_Y | 0x078007,    // ADC2 - 16 bit ADC input (DC Tach value)
      // -----------------------  Motor2:
      PMAC_VB_Y | 0x000109,    // position scale factor (Ixx08)
      PMAC_VB_Y | 0x00011c,    // vel loop scale factor (Ixx09)
      PMAC_VB_D | 0x000108,    // Motor2 present desired position
      PMAC_VB_X | 0x000112,    // Motor2 PID D term (Ixx31)
      PMAC_VB_Y | 0x000119,    // Motor2 PID P term (Ixx30)
      PMAC_VB_X | 0x00011f,    // Motor2 PID I term (Ixx33)
      PMAC_VB_Y | 0x07800E,    // ADC3 - 16 bit ADC input (DC Tach value)
      PMAC_VB_Y | 0x07800F,    // ADC4 - 16 bit ADC input (DC Tach value)
      // -----------------------  Misc Data:
   };

   /* disable VB updates: set I55 to 0 */
   strxcpy( buf, "I55=0 ", sizeof(buf));
   rc = pmac_reg_cmd( fd, buf, buf, sizeof(buf));
   if( rc != ERR_NONE)
      return ERR_DEVICE_IO;

   /* setup background Buffer Control control variables */
   memset( &vbc, 0, sizeof(vbc));
   vbc.read_buf_size    = PM_VB_READ_BUF_NITEMS;
   vbc.read_buf_offset  = PM_VB_READ_BUF_OFFSET;
   vbc.write_buf_size   = PM_VB_WRITE_BUF_NITEMS;
   vbc.write_buf_offset = PM_VB_WRITE_BUF_OFFSET;
   rc = pmac_mem_putblk( fd, PMAC_VB_CNTL_ADDR, &vbc, sizeof(vbc));
   if( rc != ERR_NONE)
      return ERR_DEVICE_IO;

   /* send read buffer setup data to pmac */
   pmac_mem_putblk( fd, PMAC_VB_DATA_ADDR+(PM_VB_READ_BUF_OFFSET*4), 
      vb_read_setup, sizeof(vb_read_setup));
   if( rc != ERR_NONE)
      return ERR_DEVICE_IO;

   /* enable VB updates: set I55 to 1 */
   strxcpy( buf, "I55=1 ", sizeof(buf));
   rc = pmac_reg_cmd( fd, buf, buf, sizeof(buf));
   if( rc != ERR_NONE)
      return ERR_DEVICE_IO;

   return ERR_NONE;
}

/***********************************************************************/
/*               pm_mi - motor info functions                          */
/***********************************************************************/

/*----------------------------------------------------------------------
**  pm_mi_init_defaults() - Init pm_minfo_t with default values 
**     mostly for scaling variable. The mi_get_VB may update these
**     variables if the pmac values change.
**     Initialize pm_minfo_t when allocating storage. 
**----------------------------------------------------------------------
*/
void pm_mi_init_defaults( struct pm_minfo_t *m )
{
   m->I10   = I10_DEFAULT;
   m->Ixx08 = Ixx08_DEFAULT;
   m->Ixx09 = Ixx09_DEFAULT;
}

/*----------------------------------------------------------------------
**  pm_mi_get_all() - Updates the motor information structure for m1 & m2
**     can take upto 0.5 ms - need to wait on VB data refresh.
**----------------------------------------------------------------------
*/
int pm_mi_get_all( 
   int fd,                   // I: pmac driver file descriptor
   struct pm_minfo_t *m1,  // O: writes data for motor 1 here 
   struct pm_minfo_t *m2   // O: writes data for motor 2 here
)
{
   int rc;
   ushort read_cntl;
   double esec;
   struct timeval s, e;

   gettimeofday( &s, NULL);

   /* clear vb.read_cntl bit to tell PMAC to refresh data */
   read_cntl = 0;
   pmac_mem_putblk( fd, PMAC_VB_CNTL_ADDR, &read_cntl, sizeof(read_cntl));

   /* wait until vb.read_cntl == 0x01 after update, timeout after 600 usec */
   read_cntl = 0;
   m1->vb_wait_cnt = m2->vb_wait_cnt = 0;
   m1->vb_wait_sec = m2->vb_wait_sec = 0;
   while( (read_cntl==0) && (esec < 0.001000) )
   {
      m1->vb_wait_cnt++;
      m2->vb_wait_cnt++;
      pmac_mem_getblk( fd, &read_cntl, PMAC_VB_CNTL_ADDR, sizeof(read_cntl));
      //printf(" read_cntl %d cnt %d esec %.6f \n", read_cntl, m1->vb_wait_cnt, esec );
      if( read_cntl)
         break;
      gettimeofday( &e, NULL);
      esec = pm_elapse_sec( &s, &e); 
   }

   // data not updated, something wrong...VB updates not enabled?
   if( read_cntl==0 )      
      return ERR_DEVICE_IO;

   /* read data from VB */
   rc = pm_mi_get_VB( fd, m1, m2 );
   if( rc != ERR_NONE)
      return rc;

   /* get data from motor MRB */
   rc = pm_mi_get_MRB( fd, m1, m2 );
   if( rc != ERR_NONE)
      return rc;

   /* get data from OpenBuffer plc0 */
   rc = pm_mi_get_plc0( fd, m1, m2 );
   if( rc != ERR_NONE)
      return rc;

   gettimeofday( &e, NULL);                  // time stamp this procedures
   m1->vb_wait_sec = m2->vb_wait_sec = pm_elapse_sec( &s, &e); 
   return ERR_NONE;
}

/*----------------------------------------------------------------------
**  pm_mi_get_MRB() - Updates the motor information structure
**     with data from the MRB buffers. 
**----------------------------------------------------------------------
*/
int pm_mi_get_MRB( 
   int fd,                   // I: pmac driver file descriptor
   struct pm_minfo_t *m1,  // O: writes data for motor 1 here 
   struct pm_minfo_t *m2   // O: writes data for motor 2 here
)
{
   int rc;
   long timestamp;
   struct pmac_MRB_raw_t md[2];

   /* get raw data from Motor Reporting DPR */

   rc = pmac_mem_get_motor_data( fd, &md, &timestamp, PMAC_DPR_MOTOR0, 2);
   if( rc != ERR_NONE)
      return ERR_DEVICE_IO;

   /* copy/convert data from MotorReportingBuffer to minfo_t */

   m1->timestamp_mrb = timestamp;
   m1->sstatus = md[0].sstatus;
   m1->mstatus = md[0].mstatus;
   m1->pdac    = dprShortFixed( md[0].pdac ) >> 8;   // 24 bits to 16 bits.
   m1->mpos    = dprLongFixed( md[0].hwpos, 1.0/(m1->Ixx08*32.0));
   m1->apos    = dprLongFixed( md[0].apos, 1.0/(m1->Ixx08*32.0) );
   m1->bpos    = dprLongFixed( md[0].bpos, 1.0/(m1->Ixx08*32.0)  );
   m1->pferr   = dprLongFixed( md[0].pepos, 1.0/(m1->Ixx08*32.0)  );
   m1->fvel    = dprShortFixed( md[0].fvel ) / (m1->Ixx09*32.0) * (8388608.0/m1->I10) * 1000.0;

   m2->timestamp_mrb = timestamp;
   m2->sstatus = md[1].sstatus;
   m2->mstatus = md[1].mstatus;
   m2->pdac    = dprShortFixed( md[1].pdac ) >> 8;   // 24 bits to 16 bits.
   m2->mpos    = dprLongFixed( md[1].hwpos, 1.0/(m2->Ixx08*32.0));
   m2->apos    = dprLongFixed( md[1].apos, 1.0/(m2->Ixx08*32.0) );
   m2->bpos    = dprLongFixed( md[1].bpos, 1.0/(m2->Ixx08*32.0)  );
   m2->pferr   = dprLongFixed( md[1].pepos, 1.0/(m2->Ixx08*32.0)  );
   m2->fvel    = dprShortFixed( md[1].fvel ) / (m2->Ixx09*32.0) * (8388608.0/m2->I10) * 1000.0;

   return ERR_NONE;
}

/*----------------------------------------------------------------------
**  pm_mi_get_VB() - Reads and converts data from the read Variable 
**     Buffer.
**----------------------------------------------------------------------
*/
int pm_mi_get_VB( 
   int fd,                   // I: pmac driver file descriptor
   struct pm_minfo_t *m1,  // O: writes data for motor 1 here 
   struct pm_minfo_t *m2   // O: writes data for motor 2 here
)
{
   int rc, addr;
   struct pmac_vb_cntl_t vbc;
   struct pm_VB_read_raw_t vbrd;

   /* get vb control memory block */
   rc = pmac_mem_getblk( fd, &vbc, PMAC_VB_CNTL_ADDR, sizeof(vbc));

   /* get vb read data */
   addr = PMAC_VB_DATA_ADDR + (PM_VB_READ_BUF_OFFSET*4) + (PM_VB_READ_BUF_NITEMS*4);
   rc = pmac_mem_getblk( fd, &vbrd, addr, sizeof(vbrd));

   /* convert/copy raw vb read data to m1 */
   m1->timestamp_vb = vbc.timestamp;
   m1->read_cntl = vbc.read_cntl;
   m1->Ixx08     = dprShortFixed( vbrd.m1.Ixx08 );
   m1->Ixx09     = dprShortFixed( vbrd.m1.Ixx09 );
   m1->dpos      = dprLongFixed( vbrd.m1.dpos, 1.0/(m1->Ixx08*32.0));
   m1->pid_d     = dprShortFixed( vbrd.m1.pid_d );
   m1->pid_p     = dprShortFixed( vbrd.m1.pid_p );
   m1->pid_i     = dprShortFixed( vbrd.m1.pid_i );
   m1->adc1      = 0x0000ffff & (vbrd.m1.adc1 >> 8);
   m1->adc2      = 0x0000ffff & (vbrd.m1.adc2 >> 8);

   /* convert/copy raw vb read data to m2 */
   m2->timestamp_vb = vbc.timestamp;
   m2->read_cntl = vbc.read_cntl;
   m2->Ixx08     = dprShortFixed( vbrd.m2.Ixx08 );
   m2->Ixx09     = dprShortFixed( vbrd.m2.Ixx09 );
   m2->dpos      = dprLongFixed( vbrd.m2.dpos, 1.0/(m2->Ixx08*32.0));
   m2->pid_d     = dprShortFixed( vbrd.m2.pid_d );
   m2->pid_p     = dprShortFixed( vbrd.m2.pid_p );
   m2->pid_i     = dprShortFixed( vbrd.m2.pid_i );
   m2->adc1      = 0x0000ffff & (vbrd.m2.adc1 >> 8);
   m2->adc2      = 0x0000ffff & (vbrd.m2.adc2 >> 8);

   return ERR_NONE;
}

/*----------------------------------------------------------------------
**  pm_mi_get_plc0() - Reads and converts data from OpenBuffer's PLC0 variables
**----------------------------------------------------------------------
*/
int pm_mi_get_plc0( 
   int fd,                   // I: pmac driver file descriptor
   struct pm_minfo_t *m1,  // O: writes data for motor 1 here 
   struct pm_minfo_t *m2   // O: writes data for motor 2 here
)
{
   int rc,
       addr;
   struct pm_VB_plc0_raw_t raw;

   // get plc0 data -> located at begining of VB Open Buffer 
   addr = PMAC_VB_DATA_ADDR + (PM_VB_OPEN_BUF_OFFSET*4);
   rc = pmac_mem_getblk( fd, &raw, addr, sizeof(raw));

   /* convert vb read raw data to minfo_t format */

   // shared data
   m1->timestamp_plc0 = m2->timestamp_plc0 = dprShortU( raw.ts );
   m1->period_plc0    = m2->period_plc0    = dprShortFixed( raw.period );

   // m1 data
   m1->base   = dprShortFixed(raw.m1_base);
   m1->dac0   = dprShortFixed(raw.m1_dac0);
   m1->dac1   = dprShortFixed(raw.m1_dac1);
   m1->dac2   = dprShortFixed(raw.m1_dac2);

   // m2 data
   m2->base   = dprShortFixed(raw.m2_base);
   m2->dac0   = dprShortFixed(raw.m2_dac0);
   m2->dac1   = dprShortFixed(raw.m2_dac1);
   m2->dac2   = dprShortFixed(raw.m2_dac2);

   return ERR_NONE;
}

/******************************************************************
**   static pm functions.
*******************************************************************
*/

/*-----------------------------------------------------------
** pm_elapse_sec() - calulates elapse time
**-----------------------------------------------------------
*/
static double pm_elapse_sec ( struct timeval *start, struct timeval *end)
{
   double sec;
   sec = end->tv_sec - start->tv_sec;
   sec += (end->tv_usec - start->tv_usec)/1000000.0;
   return sec;
}
