/*****************************************************************
** pm.h - PM_function are higer level pmac functions that are
**    customize to the IRTF T3 project. Sits between the PMAC
**    device driver (pmac.c, pmac_nice.c) and the application.
******************************************************************
*/

#ifndef _PM_H
#define _PM_H

//----------------------------------------------------------------------------------
// Variable (read/write) Buffer related defines
// RAM at 0x1140 to 0x344C (DP$060450-DP:$060FFF) is used for the Variable Buffer
// Open Space. These defines organizates the ram in 3 area: 
//                             PC_Addr  PC_Size  DSP_Addr DSP_size
//    Variable Read Buffer  -  0x1140   0x1260   $60450   0x5d8
//    Variable Write Buffer -  0x28A0   0x0bb0   $60A2a   0x2ec
//    Open buffer           -  0x3450   0x0bb0   $60d14   0x2ec
//----------------------------------------------------------------------------------

#define PM_VB_READ_BUF_NITEMS    16       //  Number of entries in read buffer request  
#define PM_VB_READ_BUF_OFFSET    0        //  offset to Read Buffer data area (DSP words)

#define PM_VB_WRITE_BUF_NITEMS   0        //  Number of entries in read buffer request  
#define PM_VB_WRITE_BUF_OFFSET   0x5d8    //  offset to Write Buffer data area (DSP words)

#define PM_VB_OPEN_BUF_OFFSET    0x8c4    //  offset to Open Buffer data area (DSP words)

struct pm_VB_read_raw_motor_t   /* DPR Variable Read Buffer raw data  (single motor data)  */
                                /* note address are for motor1/motor2 */
{  
   long      Ixx08;           // Y:$00089/109 - Ixx08 position scale factor
   long      Ixx09;           // Y:$0009c/11c - Ixx09 velocity loop scale factor
   long      dpos[2];         // D:$00088/108 - present desired position
   long      pid_d;           // X:$00092/112 - PID D term (Ixx31)
   long      pid_p;           // Y:$00099/119 - PID P term (Ixx30)
   long      pid_i;           // X:$0009f/11f - PID I term (Ixx33)
   ulong     adc1;            // Y:$078006/$07800E - ADC1/3 16-bit analog input. bits 8-16
   ulong     adc2;            // Y:$078007/$07800F - ADC2/4 16-bit analog input. bits 8-16
}; 

struct pm_VB_read_raw_t       /* DPR Variable Buffer raw data - has data for 2 motors */
{  
   struct pm_VB_read_raw_motor_t m1;      // data from motor 1
   struct pm_VB_read_raw_motor_t m2;      // data from motor 2
}; 
   
struct pm_VB_plc0_raw_t    // DPR Variable for PLC0 - located in Open Buffer
{
   long ts;         // timestamp (servo cycle count)
   long period;     // servo cycles between PCL executions
                    // MOTOR1 - HA Axis
   long m1_base;     // backlash base value 
   long m1_dac0;     // dac output from motor PID loop
   long m1_dac1;     // dac output for motor DAC1
   long m1_dac2;     // dac output for motor DAC2
                    // MOTOR2 - DEC axis
   long m2_base;     // backlash base value 
   long m2_dac0;     // dac output from motor PID loop
   long m2_dac1;     // dac output for motor DAC1
   long m2_dac2;     // dac output for motor DAC2
};

//------------------------------------------------------------------------------
// motor info  - all data gather from the DRP interfaces
//    (Reporting Buffer & Read Buffer) 
//------------------------------------------------------------------------------
struct pm_minfo_t
{  
   //----------------------------------------------
   // various i variables needed for scaling data
   //----------------------------------------------
   double I10;      // servo interrupt time
   float Ixx08;     // position scale factor       (updated by get_VB)
   float Ixx09;     // velocity loop scale factor  (updated by get_VB)
   
   //---------------------------------------------- 
   // motor data (from MRB)
   //----------------------------------------------
   long   timestamp_mrb;     // MBR timestamp
   struct pmac_ss_t sstatus; // servo status bits
   struct pmac_ms_t mstatus; // motor status bits
   long   pdac;              // Digital value of motor DAC (represents the 16 bits value)
   double mpos;              // master position (counts)
   double apos;              // actual position (counts)
   double bpos;              // position bias (counts)
   double pferr;             // position following error (counts)
   double fvel;              // filtered/average velocity (cts/sec)
   
   //----------------------------------------------
   // motor data (from VB)
   //----------------------------------------------
   long   timestamp_vb;     // VB timestamp
   ushort read_cntl;        // VB - Read Control word.
   int vb_wait_cnt;         // number of loop until VB data was refreshed. 
   double vb_wait_sec;      // time it took to update minfo in seconds
   double dpos;             // present desired position (D:$88)
   long   pid_d;            // PID D term (Ixx31)
   long   pid_p;            // PID P term (Ixx30)
   long   pid_i;            // PID I term (Ixx33)
   long   adc1;             // Tachometer DAC1
   long   adc2;             // Tachometer DAC2

   //----------------------------------------------
   // plc0 data - dual DAC function
   //----------------------------------------------
   long   timestamp_plc0;     // timestamp
   long   period_plc0;        // servo cycles
   long   base;               // base (backlash) value
   long   dac0;               // dac output from PID (same as pdac)
   long   dac1;               // dac output for motor dac1
   long   dac2;               // dac output for motor dac2
};


/*-----------------------------------------------------
** function prototypes
**----------------------------------------------------
*/
int pm_apos_set( int fd, long ra, long dec );

int pm_init( int fd );

int pm_vb_init( int fd );

void pm_mi_init_defaults( struct pm_minfo_t *m );
int pm_mi_get_all( int fd, struct pm_minfo_t *m1, struct pm_minfo_t *m2 );
int pm_mi_get_MRB( int fd, struct pm_minfo_t *m1, struct pm_minfo_t *m2 );
int pm_mi_get_VB( int fd, struct pm_minfo_t *m1, struct pm_minfo_t *m2 );
int pm_mi_get_plc0( int fd, struct pm_minfo_t *m1, struct pm_minfo_t *m2 );



#endif  // _PM_H

