/***********************************************************************
**  Device driver for DeltaTau PMAC PCI lite motion controller board
**  Linux 2.4 kernel. 
**   - some functions based on pmacpci driver obtained from Delta Tau 
**     Technical support.
** --------------------------------------------------------------------
**  Written by:
**     Tony Denault <denault@irtf.ifa.hawaii.edu>
**
**     Institute for Astronomy, University of Hawaii
**     2680 Woodlawn Drive, Honolulu, HI 96822
**     http://www.irtfweb.ifa.hawaii.edu
************************************************************************
*/

 
#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

/*------------------------------
** include files
**------------------------------
*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
                                                                                
#include <linux/sched.h>
#include <linux/kernel.h> /* printk() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/tqueue.h>
#include <linux/types.h>    /* size_t   */
//#include <linux/malloc.h>  // replaced by slab.h in 2.4
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/pci.h>
#include <linux/delay.h>

#include <asm/uaccess.h>    /* copy from/to user */
#include <asm/io.h>

#include "pmac.h"

/*------------------------------
** defines & declare structure
**------------------------------
*/
MODULE_LICENSE("GPL");
#define DEBUG 1

#define PMAC_VERSION     "03.12"  // Version is YY.MM 
#define PMAC_MAJOR       66       // major device number

#define PMAC_PCI_VENDOR_ID 0x1172  // vendor ID for locating  PCI board
#define PMAC_PCI_DEVICE_ID 0x0001  // Devicd ID for locating  PCI board

#define IO_UDELAYS             5   // micro seconds delay per try
#define IO_TIMEOUT_CNT      1000   // micro seconds delay per try

#define CHAR_NUL    (0x00)
#define CHAR_CR     (0x0d)   // Carriage_return
#define CHAR_ACK    (0x06)  
#define CHAR_CTRL_A (0x01)  
#define CHAR_CTRL_G (0x07)  
#define CHAR_CTRL_X (0x18)  
#define CHAR_CTRL_Z (0x1a)  

#define PT_UNKNOWN 0 
#define PT_PMAC1   1  // PMAC1
#define PT_PMAC2   2  // PMAC2
#define PT_PMACUL  3  // PMACU
#define PT_PMAC    4  //
#define PT_PMAC1T  5  // TURBO1
#define PT_PMAC2T  6  // TURBO2
#define PT_PMACUT  7  // TURBOU
#define PT_UMAC    8 
#define PT_QMAC    9

#define MAX_COMM_BUFFER  256   // max size of communication buffer( register IO and RPR reads)
#define MAX_MEM_PUTS     160   // max size for DPR write line 

#define MAX( a, b )      ((a) < (b) ? (b) : (a))
#define MIN( a, b )      ((a) < (b) ? (a) : (b))

struct stuff_t
{
  struct semaphore sem;           // mutual exclusion semaphore 

  struct pci_dev *pdev;           // PMAC's PCI device structure
  unsigned long io_base;          // PMAC's PCI IO port address start/end
  unsigned long io_base_end;   
  int  irq;                       // PMAC's IRQ value.
  unsigned long dpram;            // PMAC's PCI Dual Ported RAM address start/end
  unsigned long dpram_end;   

  int  dpr_enable;                // is access to DPR possible
  char * dpr_vadd;                // virtual mapped address for Dual Port RAM  (base address)
  char * reg_write_status;        // host output control register address
  char * reg_read_status;         // host input control register address
  char * reg_cntl_char;           // host control char write register address
  char * reg_num_char;            // number of chars in string register address
  char * reg_write;               // write data register address
  char * reg_read;                // read  data register address

  int    type;                    // type of PMAC PCI board, see PT_* 
  char   type_str[20];
  int    num_procs;               // number of processes with device open 

};

/*-----------------------------
** prototype functions
**-----------------------------
*/
int  init_module(void);
void cleanup_module (void);

static int pmac_open(struct inode *inode, struct file *filp);
static int pmac_release(struct inode *inode, struct file *filp);
static int pmac_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg);

static int ioctl_string_reverse( char * buf, int strlen );
static int ioctl_reg_cmd( char * cmd, int cmd_len, char * rpy, int  rpy_size, struct stuff_t *s);
static int ioctl_mem_puts( char * cmd, int  cmd_len, struct stuff_t *s);
static int ioctl_mem_gets( char * rpy, int  rpy_size, struct stuff_t *s);
static int ioctl_mem_cmd( char * cmd, int cmd_len, char * rpy, int  rpy_size, struct stuff_t *s);
static int ioctl_mem_getblk( char * dest, int src, int nbytes, struct stuff_t * s);
static int ioctl_mem_putblk( int dest, char * src, int nbytes, struct stuff_t *s );
static int ioctl_mem_get_motor_data( char * dest, long* timestamp, 
                                     int m0, int nmotors, struct stuff_t * s);
static int ioctl_t1( struct pmac_parms_t *p, struct stuff_t *s );
static int ioctl_t2( struct pmac_parms_t *p, struct stuff_t *s );

static int pmac_detect_pci( struct stuff_t * s );
static int pmac_init_type_dpram( struct stuff_t * s );

static int io_write_char( char outchar, struct stuff_t *s );
static int io_write_line( char *buffer, struct stuff_t *s);
static int io_read_line( char *buffer, int buffer_size, struct stuff_t *s );
static int io_flush( struct stuff_t *s );
static int io_command( char * cmd, char * rpy, int  rpy_size, struct stuff_t *s);

static int mem_write_line( char * buf, struct stuff_t *s);
static int mem_read_line( int * r_nc, char * buf, int buf_size, struct stuff_t *s );
static int mem_flush( struct stuff_t *s );

static int ms2jiffies( long ms );
static void msleep( int ms );

/*-----------------------------
** global variable
**-----------------------------
*/
struct file_operations Pmac_fops =
{
        open:           pmac_open,
        ioctl:          pmac_ioctl,
        release:        pmac_release
};
 
static struct stuff_t Stuff;                    // hold internal driver info 

/*-----------------------------------------
** Inline function for register access
**-----------------------------------------
*/


static inline int io_reg_status( int base )  { return base+2; }   // return ioaddr of status reg
static inline int io_reg_lo_init( int base ) { return base+5; }   // return ioaddr of lo_init reg
static inline int io_reg_hi_init( int base ) { return base+6; }   // return ioaddr of hi_init reg
static inline int io_reg_data( int base )    { return base+7; }   // return ioaddr of data reg

static inline int io_rdy_to_rcv( int base )  {    // pmac is ready to recieve a char
   return ((inb_p( base+2 ) & 0x02) == 0x02);
}
static inline int io_rdy_to_send( int base )  {   // pmac is ready to send char  
   return ((inb_p( base+2 ) & 0x01) == 0x01);
}

static inline int dpr_read_ready( struct stuff_t *s ) { return readw(s->reg_read_status) != 0; }   
static inline int dpr_read_error( struct stuff_t *s ) { return (readw(s->reg_read_status) & 0x8000) > 0; }   
static inline int dpr_read_error_code( struct stuff_t *s ) { return readw(s->reg_read_status) & 0x00fff; }   
static inline int dpr_read_ack( struct stuff_t *s ) { return (readw(s->reg_read_status) & 0x00ff) == CHAR_ACK; }   
static inline void dpr_clr_read_ready( struct stuff_t *s ) { writew(0, s->reg_read_status); }   
static inline int dpr_write_ready( struct stuff_t *s ) { return (readw(s->reg_write_status) & 0x0001) == 0; }   
static inline void dpr_set_write_ready( struct stuff_t *s ) { writew(1, s->reg_write_status); }   

/*-----------------------------
** init_module()
**-----------------------------
*/
int init_module(void)
{
   int rc;

   printk( KERN_INFO "pmac: DeltaTau PMAC PCI Lite Motion Controller Driver %s\n", PMAC_VERSION);
   SET_MODULE_OWNER(&Pmac_fops); 

   /* initialize Stuff  - internal driver data structure */
   memset( &Stuff, 0, sizeof(Stuff) );    // clear device data structure
   sema_init( &Stuff.sem, 1);             // initialize driver mutex

   /*------------------------------
   ** Find the PCI device. 
   */
   if( pmac_detect_pci( &Stuff ) )
   {
      printk( KERN_ERR  "pmac: pmac_detect_pci() error \n" );
      return -ENODEV;
   }

   /*------------------------------
   ** finish pmac initialization
   */
   outb_p( 0, io_reg_lo_init(Stuff.io_base)); 
   outb_p( 0, io_reg_hi_init(Stuff.io_base)); 
   
   rc = io_flush( &Stuff );
   if( rc < 0 )
   {
      printk( KERN_ERR  "pmac: io_flush() Initialization failure \n" );
      return -ENODEV;
   }

   pmac_init_type_dpram( &Stuff );
   if( rc < 0 )
   {
      printk( KERN_ERR  "pmac: pmac_init_type_dpram() error.\n" );
      return -ENODEV;
   }


   /*------------------------------
   ** Register the device. 
   */

   printk( KERN_INFO "pmac: MPAC found: IO:0x%lx-%lx DPRAM:0x%lx-%lx IRQ:%d Type:%d\n", 
      Stuff.io_base, Stuff.io_base_end, 
      Stuff.dpram, Stuff.dpram_end, Stuff.irq, Stuff.type );

   /*------------------------------
   ** Register the device. 
   */
   rc = register_chrdev( PMAC_MAJOR, "pmac", &Pmac_fops);
   if (rc != 0)
   {
      printk(KERN_ERR "pmac: unable to get major %d\n",  PMAC_MAJOR);
      return rc;
   }

   return 0;
}
 
/*-----------------------------
** cleanup_module()
**-----------------------------
*/
void cleanup_module(void)
{
   // We free the major number 
   unregister_chrdev (PMAC_MAJOR, "pmac");

   /* We free our buffer pmac */

   //printk( KERN_DEBUG "pmac: cleanup_module()\n");
}

/*-----------------------------
** pmac_open()
**-----------------------------
*/
static int pmac_open(struct inode *inode, struct file *filp)
{
   MOD_INC_USE_COUNT;
   Stuff.num_procs++;

   //printk( KERN_DEBUG "pmac: pmac_open()\n");
   return 0;
}
 
/*-----------------------------
** pmac_release()
**-----------------------------
*/
static int pmac_release(struct inode *inode, struct file *filp)
{
   MOD_DEC_USE_COUNT;
   Stuff.num_procs--;

   // printk( KERN_DEBUG "pmac: pmac_release()\n");
   return 0;
}

/*-----------------------------
** pmac_ioctl()
**-----------------------------
*/
static int pmac_ioctl(struct inode *inode, struct file *filp,
                unsigned int cmd, unsigned long arg)
{
   struct pmac_parms_t p;          // local copy of 'arg'

   copy_from_user( &p, (void*)arg, sizeof(p));

   switch( cmd )
   {
      case PMAC_IOCTL_MSLEEP:
      {
         p.ret = 0;
         down( &Stuff.sem);
         msleep( (int) p.l[0] );
         up( &Stuff.sem);
         break;
      }

      case PMAC_IOCTL_STRING_REVERSE:
      {
         p.ret = 0;
         down( &Stuff.sem);
         ioctl_string_reverse( p.p[0], p.l[0] );
         up( &Stuff.sem);
         break;
      }

      case PMAC_IOCTL_REG_CMD:
      {
         down( &Stuff.sem);
         p.ret = ioctl_reg_cmd( p.p[0], p.l[0], p.p[1], p.l[1], &Stuff ); 
         up( &Stuff.sem);
         break;
      }

      case PMAC_IOCTL_MEM_PUTS:
      {
         down( &Stuff.sem);
         p.ret = ioctl_mem_puts( p.p[0], p.l[0], &Stuff ); 
         up( &Stuff.sem);
         break;
      }

      case PMAC_IOCTL_MEM_GETS:
      {
         down( &Stuff.sem);
         p.ret = ioctl_mem_gets( p.p[0], p.l[0], &Stuff ); 
         up( &Stuff.sem);
         break;
      }

      case PMAC_IOCTL_MEM_CMD:
      {
         down( &Stuff.sem);
         p.ret = ioctl_mem_cmd( p.p[0], p.l[0], p.p[1], p.l[1], &Stuff ); 
         up( &Stuff.sem);
         break;
      }

      case PMAC_IOCTL_MEM_FLUSH:
      {
         down( &Stuff.sem);
         p.ret = mem_flush( &Stuff ); 
         up( &Stuff.sem);
         break;
      }

      case PMAC_IOCTL_MEM_GETBLK:
      {
         down( &Stuff.sem);
         p.ret = ioctl_mem_getblk( p.p[0], p.l[0], p.l[1], &Stuff ); 
         up( &Stuff.sem);
         break;
      }

      case PMAC_IOCTL_MEM_GET_MOTOR_DATA:
      {
         down( &Stuff.sem);
         p.ret = ioctl_mem_get_motor_data( p.p[0], &p.l[2], p.l[0], p.l[1], &Stuff ); 
         up( &Stuff.sem);
         break;
      }

      case PMAC_IOCTL_MEM_PUTBLK:
      {
         down( &Stuff.sem);
         p.ret = ioctl_mem_putblk( p.l[0],  p.p[0], p.l[1], &Stuff );
         up( &Stuff.sem);
         break;
      }

      case PMAC_IOCTL_T1:
      {
         down( &Stuff.sem);
         p.ret = ioctl_t1( &p, &Stuff ); 
         up( &Stuff.sem);
         break;
      }

      case PMAC_IOCTL_T2:
      {
         down( &Stuff.sem);
         p.ret = ioctl_t2( &p, &Stuff ); 
         up( &Stuff.sem);
         break;
      }

      default:
         p.ret = -EFAULT;
         if(DEBUG) printk( KERN_DEBUG "pmac: received unknown command (%d)\n",cmd );
         break;
   }

   copy_to_user( (void*)arg, &p, sizeof(p));
   return 0;
}    

/***********************************************************************************/
/**  Level1 PMAC function - ioctl functions                                       **/
/***********************************************************************************/

/*-------------------------------------------------------------------------
** ioctl_string_reverse() - reverse buf, user space string.
**-------------------------------------------------------------------------
*/
static int ioctl_string_reverse( 
   char * buf,  // string to reverse.
   int strlen   // strlen(buf)
)
{
   char ca, cb;
   int a, b;

   a = 0;
   b = strlen-1;

   while( a < b )
   {
      get_user( ca, &buf[a]);  // ca = buf[a];
      get_user( cb, &buf[b]);  //cb = buf[b];

      put_user( cb, &buf[a]);   // buf[a] = cb;
      put_user( ca, &buf[b]);   // buf[b] = ca;

      a++;
      b--;
   }
   return 0;
}

/*-------------------------------------------------------------------------
** ioctl_reg_cmd() - pmac register-based command IO
**-------------------------------------------------------------------------
*/
static int ioctl_reg_cmd( 
   char * cmd,          // I: pmac command - user space string address
   int  cmd_len,        // I: strlen(cmd)
   char * rpy,          // O: pmac reply - user space string address
   int  rpy_size,       // I: sizeof(rpy)
   struct stuff_t *s    // driver data
)
{
   int  l;
   char buffer[MAX_COMM_BUFFER];

   // copy command into kernel
   l = MIN( cmd_len, sizeof(buffer)-1);
   copy_from_user( buffer, cmd, l);
   buffer[l] = 0;

   io_flush( s );
   io_write_line( buffer, s );
   l = io_read_line( buffer, sizeof(buffer), s );

   // copy reply to user          
   if( l > 0 ) 
   {
      l = MIN( l, rpy_size-1);
      buffer[l] = 0;
      copy_to_user( rpy, buffer, l+1);
   }
   else
      put_user( CHAR_NUL, rpy);  // reply contains ""

   return 0;
}

/*-------------------------------------------------------------------------
** ioctl_mem_puts() - write command to DPR interface
**-------------------------------------------------------------------------
*/
static int ioctl_mem_puts( 
   char * cmd,          // I: pmac command - user space string address
   int  cmd_len,        // I: strlen(cmd)
   struct stuff_t *s    // driver data
)
{
   int  rc, l;
   char buffer[MAX_MEM_PUTS];

   // copy command into kernel
   l = MIN( cmd_len, sizeof(buffer)-1);
   copy_from_user( buffer, cmd, l);
   buffer[l] = 0;

   rc = mem_write_line( buffer, s );
   return rc;
}

/*-------------------------------------------------------------------------
** ioctl_mem_gets() - reads a reply from DPR interface
**-------------------------------------------------------------------------
*/
static int ioctl_mem_gets( 
   char * rpy,          // O: pmac reply - user space string address
   int  rpy_size,       // I: sizeof(rpy)
   struct stuff_t *s    // driver data
)
{
   int  rc, nc;
   char buffer[MAX_COMM_BUFFER];

   rc = mem_read_line( &nc, buffer, sizeof(buffer), s );
   //printk( KERN_DEBUG "mem_readline()=%d  buffer=%s status 0x%x\n", rc, buffer, readw(s->reg_read_status));

   // copy reply to user          
   if( (nc > 0) && (rpy_size > 1)) 
   {
      nc = MIN( nc, rpy_size-1);
      buffer[nc] = 0;
      copy_to_user( rpy, buffer, nc+1);
   }
   else
      put_user( CHAR_NUL, rpy);  // reply contains ""

   return rc;
}

/*-------------------------------------------------------------------------
** ioctl_mem_cmd() - does put/get combination. The get may be called
**    multiple times.
**-------------------------------------------------------------------------
*/
static int ioctl_mem_cmd( 
   char * cmd,          // I: pmac command - user space string address
   int  cmd_len,        // I: strlen(cmd)
   char * rpy,          // O: pmac reply - user space string address
   int  rpy_size,       // I: sizeof(rpy)
   struct stuff_t *s    // driver data
)
{
   int  rc, 
        bufsize,
        t_nc, nc;
   char buffer[MAX_COMM_BUFFER];

   /*
   ** write command
   */
   rc = ioctl_mem_puts( cmd, cmd_len, s );
   if( rc ) 
      return rc;

   /*
   ** Read response (may call read line multiple times).
   */
   udelay( 1 );
   bufsize = sizeof(buffer);
   t_nc = 0;
   rc = -EAGAIN;
   while( rc == -EAGAIN )
   {
      rc = mem_read_line( &nc, &buffer[t_nc], bufsize, s );
      bufsize -= nc;
      t_nc += nc;
      //printk( KERN_DEBUG "mem_cmd: rc=%d nc=%d \n", rc, nc);
   }

   // copy reply to user          
   if( (t_nc > 0) && (rpy_size > 1)) 
   {
      t_nc = MIN( t_nc, rpy_size-1);
      buffer[t_nc] = 0;
      copy_to_user( rpy, buffer, t_nc+1);
   }
   else
      put_user( CHAR_NUL, rpy);  // reply contains ""


   return rc;
}

/*-------------------------------------------------------------------------
** ioctl_mem_getblk() - copies a block of DRP to the caller
**-------------------------------------------------------------------------
*/
static int ioctl_mem_getblk( 
   char * dest,         // O: dest - user space string address
   int    src,          // I: src - beg IO address to copy (offset from base).
   int    nbytes,       // I: number of bytes to copy
   struct stuff_t *s    // driver data
)
{
   char *sptr;

   sptr = s->dpr_vadd + src;   // add PCI base offset to src.

   while( nbytes-- )
   {
      put_user( readb(sptr), dest); 
      dest++;
      sptr++;
   }
   return 0;
}

/*-------------------------------------------------------------------------
** ioctl_mem_putblk() - copies user data to pmac's IO MEM
**-------------------------------------------------------------------------
*/
static int ioctl_mem_putblk( 
   int    dest,         // O: dest - IO address to copy to         
   char * src,          // I: src - user space address to copy from
   int    nbytes,       // I: number of bytes to copy
   struct stuff_t *s    // driver data
)
{
   char *dptr;
   unsigned char c;

   dptr = s->dpr_vadd + dest;  // dptr is kernel vir addess for IO memory.

   //printk( KERN_DEBUG "ioctl_mem_putblk(): D:%d %lx  nbyte:%d \n", dest, (long)dptr, nbytes );
   while( nbytes-- )
   {
      get_user( c, src );  // get user data
      writeb( c, dptr);    // Write to IO         
      dptr++;
      src++;
   }
   return 0;
}

/*-------------------------------------------------------------------------
** ioctl_mem_get_motor_data() - copies motor data (and timestamp) from DPR to user memory
**-------------------------------------------------------------------------
*/
static int ioctl_mem_get_motor_data( 
   char * dest,          // O: dest - copy data to this user space buffer
   long * timestamp,     // O: timestamp from DPR
   int m0,               // I: begining offset for motor 0 = PMAC_DPR_MOTOR0, ...
   int nmotors,          // I: number of motor block to copy
   struct stuff_t * s    // driver data
)
{
   int nbytes;
   char *sptr;

   /* copy get timestamp */
   *timestamp = readl(s->dpr_vadd+0x0678) & 0x7fffffff;

   /* copy motor data to caller */
   sptr = s->dpr_vadd + m0;   // add PCI base offset to src.
   nbytes = PMAC_DPR_MOTOR_NBYTES * nmotors;
   while( nbytes-- )
   {
      put_user( readb(sptr), dest); 
      dest++;
      sptr++;
   }
   return 0;
}

/*-------------------------------------------------------------------------
** ioctl_t1() - pmac register-based command IO
**-------------------------------------------------------------------------
*/
static int ioctl_t1( 
   struct pmac_parms_t *p,  // copy of 'arg'
   struct stuff_t *s        // driver data
)
{
   int write, read, nchar;

   read = readw(s->reg_read_status);
   write = readw(s->reg_write_status);
   nchar = readw(s->reg_num_char);

   printk( KERN_DEBUG "pmac: t1 write=0x%x read=0x%x nchar=%d \n", write, read, nchar );
   printk( KERN_DEBUG "pmac: t1 read 0x%x 0x%x 0x%x 0x%x  \n",
      readw(s->reg_read), readw(s->reg_read+1), readw(s->reg_read+2), readw(s->reg_read+3));
   
   return 0;
}

/*-------------------------------------------------------------------------
** ioctl_t2() - pmac register-based command IO
**-------------------------------------------------------------------------
*/
static int ioctl_t2( 
   struct pmac_parms_t *p,   // copy of 'arg'
   struct stuff_t *s         // driver data
)
{
   return 0;
}


/***********************************************************************************/
/**  Level0 PMAC function - lower level hardware/device driver functions          **/
/***********************************************************************************/

/*---------------------------------------------------------------------
**  pmac_detect_pci() - detect and fill s with PCI device information
**     should be called only by driver _init() function.
**---------------------------------------------------------------------
*/
static int pmac_detect_pci( struct stuff_t * s )
{
   int i;
   unsigned char myirq;
   unsigned long start, end, flags;

   /* find pci device */
   s->pdev = pci_find_device( PMAC_PCI_VENDOR_ID, PMAC_PCI_DEVICE_ID, NULL);
   if( s->pdev == NULL )
      return -ENODEV;

   /* located the IO and DualPort RAM locations, and the irq value */
   for( i=0; i<6; i++ )
   {
      start = pci_resource_start( s->pdev, i );
      end   = pci_resource_end( s->pdev, i );
      flags   = pci_resource_flags( s->pdev, i );

      if( flags & IORESOURCE_IO )
      {
         s->io_base      = start;
         s->io_base_end  = end;
      }
      if( flags & IORESOURCE_MEM )
      {
         s->dpram       = start;
         s->dpram_end   = end;
      }
   }

   /* save our IRQ number */
   pci_read_config_byte( s->pdev, PCI_INTERRUPT_LINE, &myirq );
   s->irq = myirq;

   // ioport & dpram are required
   if( ((s->io_base_end - s->io_base) < 0xff )   ||
       ((s->dpram_end  - s->dpram )  < 0x3fff ) )
   {
      return -ENODEV;
   }

   /* Everything OK, enable device. */
   pci_enable_device(s->pdev); 

   return 0;
}

/*---------------------------------------------------------------------
**  pmac_init_type_dpram() - detect and fill the s->type. Also initialize
**     the DPRAM address (whose values  depend on type)
**---------------------------------------------------------------------
*/
static int pmac_init_type_dpram( struct stuff_t * s )
{
   int rc;
   char buf[40];

    rc = io_command( "type\0", s->type_str, sizeof(s->type_str), s );
    if( rc < 0 )
       return rc;
   //printk( KERN_DEBUG "pmac: detect_type = %s\n", s->type_str);

   if(strstr(s->type_str, "PMAC1"))
      s->type = PT_PMAC1;
   else if(strstr(s->type_str, "PMAC2"))
      s->type = PT_PMAC2;
   else if(strstr(s->type_str, "PMACU"))
      s->type = PT_PMACUL;
   else if(strstr(s->type_str, "TURBO1"))
      s->type = PT_PMAC1T;
   else if(strstr(s->type_str, "TURBO2"))
      s->type = PT_PMAC2T;
   else if(strstr(s->type_str, "TURBOU"))
      s->type = PT_PMACUT;


   /*----------------------------------------------------
   ** the type is need to finish the DPR configuration.
   **----------------------------------------------------
   */

   /* enable and get dpram's virtual address */
   if( s->dpram > 0  )
   {
      s->dpr_enable = 1;
      s->dpr_vadd = ioremap( s->dpram, s->dpram_end-s->dpram+1);

      /* initialize registers pointers in DPR*/
      if( s->type > PT_PMAC)                      /* Is it Turbo! */
      {
         s->reg_write_status = s->dpr_vadd + 0x0E9C;
         s->reg_read_status  = s->dpr_vadd + 0x0F40;
         s->reg_cntl_char    = s->dpr_vadd + 0x0E9E; 
         s->reg_num_char     = s->dpr_vadd + 0x0F42;
         s->reg_write        = s->dpr_vadd + 0x0EA0;
         s->reg_read         = s->dpr_vadd + 0x0F44;
      }
      else
      {
         s->reg_write_status = s->dpr_vadd + 0x062C;
         s->reg_read_status  = s->dpr_vadd + 0x06D0;
         s->reg_cntl_char    = s->dpr_vadd + 0x062E;
         s->reg_num_char     = s->dpr_vadd + 0x06D2;
         s->reg_write        = s->dpr_vadd + 0x0630;
         s->reg_read         = s->dpr_vadd + 0x06D4;
      }

      io_command( "I58=1 \0", buf, sizeof(buf), s ); // enable DPRAM communications
      mem_flush( s );                                // flush the mem command buffer 
   }

   return 0;
}

/*-------------------------------------------------------------------
** io_write_char() - write a char to the IO output register
**-------------------------------------------------------------------
*/
static int io_write_char( 
   char outchar,
   struct stuff_t *s        // driver data
)
{
   unsigned long timeout = 0;
   while( timeout++ < IO_TIMEOUT_CNT )
   {
      if( io_rdy_to_rcv(s->io_base)) 
      {
         outb_p( outchar, io_reg_data(s->io_base));
         return 0;
      }
      else
         udelay( IO_UDELAYS );
   }
   return -EIO;
}

/*-------------------------------------------------------------------
** io_write_line() - writes a null terminated string (max MAX_COMM_BUFFER char)
**   to the data port. Write CD to end pmac command. 
** Returns < 0 - error
**         otherwise number of chars written.
**-------------------------------------------------------------------
*/

static int io_write_line( 
   char *buffer,            // buffer to write
   struct stuff_t *s        // driver data
)
{
   unsigned int nc=0;
   int rc=0;

   // output line 
   while( buffer[nc]!=CHAR_NUL && nc<(MAX_COMM_BUFFER-1) && rc==0)
      rc=io_write_char( buffer[nc++], s);

   // send carriage return
   if(rc==0)
      rc=io_write_char( CHAR_CR, s);    

   //printk( KERN_DEBUG "pmac: io_write_line() rc=%d nc=%d\n", rc, nc);
   return ( rc < 0 ? rc : nc);
}

/*-------------------------------------------------------------------
** io_read_line() - read data from the io port. IT does not send the ACK
**    character back. It checks for watchdog or junk, terminates the buffer
**    with null.
**-------------------------------------------------------------------
*/
static int io_read_line( 
   char * buffer,           // read into this buffer 
   int buffer_size,         // sizeof(buffer)
   struct stuff_t *s        // driver data
)
{
   unsigned long timeout = 0;
   int nc = 0;
   int rc = 0;

   buffer_size--; // account for NULL char

   while( (timeout<IO_TIMEOUT_CNT) && (nc<buffer_size))
   {
      if(!io_rdy_to_send(s->io_base))
      {
         udelay(IO_UDELAYS);
         timeout++;
         continue;
      }

      buffer[nc] = inb_p(io_reg_data(s->io_base));
      timeout=0;

      if(buffer[nc] == CHAR_ACK)
      {
         // end of transmission
         break;
      }
      else if(buffer[nc] == CHAR_CR )
      {
         nc++;
         break;
      }
      else if(buffer[nc] > 0x7E)
      {
         rc = -EIO;
         break;
      }
      else if(buffer[nc] == CHAR_CTRL_G) 
      {
         rc = -EIO;
         break;
      }
      else
         nc++;

      if( rc && !io_rdy_to_send(s->io_base))
         break;
   }

   if (timeout>=IO_TIMEOUT_CNT)
      rc = -EIO;

   buffer[nc] = CHAR_NUL;

   //printk( KERN_DEBUG "pmac: io_read_line() rc=%d nc=%d timeout=%ld\n", rc, nc, timeout);
   return ( rc < 0 ? rc : nc);  // less that 0 is error, or number of chars
}

/*---------------------------------------------------------------------------------
** io_flush() - clears out any garbage in pmac command buffer by writing Ctrl-X and
**    then reads few characters back to clear out data buffer.
**---------------------------------------------------------------------------------
*/
static int io_flush( struct stuff_t *s )
{
   int i=0, nchr=0;
   unsigned int temp;

   io_write_char( CHAR_CTRL_X, s);

   while(i++<IO_TIMEOUT_CNT && nchr<5)
   {
      if(!io_rdy_to_send(s->io_base))
      {
         udelay(IO_UDELAYS);
         continue;
      }
      else
         nchr++;

      temp = inb_p(io_reg_data(s->io_base));
      if(temp == CHAR_CTRL_X)
         break;
      else if(temp == 0xFF)
         return -ENODEV;
   }
   return 0;
}

/*-------------------------------------------------------------------------
** io_command() - pmac command IO using register interface.
**-------------------------------------------------------------------------
*/
static int io_command( 
   char * cmd,          // I: pmac command - kernel space address.
   char * rpy,          // O: pmac reply -  kernel space address.
   int  rpy_size,       // I: sizeof(rpy).
   struct stuff_t *s    // driver data
)
{
   io_flush( s );
   io_write_line( cmd, s );
   io_read_line( rpy, rpy_size, s );
   return 0;
}

/*-------------------------------------------------------------------
** mem_write_line() - write a ascii line to the pmac using the DPR interface 
**                    buf required to be \0 terminated. No CR. Max is 159 chars.
**-------------------------------------------------------------------
*/
static int mem_write_line( 
   char * buf,
   struct stuff_t *s        // driver data
)
{
   int i;
   char * dest;
   unsigned long timeout = 0;

   if( !s->dpr_enable )
      return -EPERM;

   dest = s->reg_write;
   while( timeout++ < IO_TIMEOUT_CNT )
   {
      if( dpr_write_ready(s) ) 
      {
         // copy buf[] to reg_write[]
         i = 0;
         while((i<=159) && (buf[i]!=0))
            writeb( buf[i++], dest++);
         
         writeb( CHAR_NUL, dest++); // terminate string with NULL

         dpr_set_write_ready(s);     // set write_status ready

         //printk( KERN_DEBUG "pmac: mem_write_line() wrote %d %s \n", i, buf);
         return 0;
      }
      else
         udelay( IO_UDELAYS );
   }
   return -EIO;
}

/*-------------------------------------------------------------------
** mem_read_line() - reads a ascii reply from DPR interface 
**    returns  0 = NO error       
**    returns <0 = Error
**                   Can be EAGAIN - means more data to read.
**-------------------------------------------------------------------
*/
static int mem_read_line( 
   int            * r_nc,     // 0: return number of chars read.
   char           * buf,      // O: store data from pmac here
   int            buf_size,   // I: sizeof( buf )
   struct stuff_t *s          // I: driver data
)
{
   int rc, nc, status;
   unsigned long timeout;

   if( !s->dpr_enable )
      return -EPERM;

   buf[0] = 0;   // null reply
   *r_nc = 0;
 
   /* wait until read_staus > 0 */
   timeout = 0;
   while( 1 )
   {
      if( dpr_read_ready(s) ) break;
      if( timeout++ > IO_TIMEOUT_CNT )
         return -ETIMEDOUT;
      udelay( IO_UDELAYS );
   }

   /* If error, not responds is available */
   if(  dpr_read_error(s)  )
   {
      //printk( KERN_DEBUG "pmac: mem_read_line() read_error_code is=%d\n", dpr_read_error_code(s));
      dpr_clr_read_ready(s);
      return -ENODATA;
   }

   /* read response string */
   nc = readw(s->reg_num_char) - 1;
   if( nc < 0 ) nc = 0;
   if( nc > 0 )
   {
      if( nc >= buf_size ) nc = buf_size-1;
      memcpy_fromio( buf, s->reg_read, nc );
   }

   /* check termination characters */
   status = readw(s->reg_read_status) & 0x00ff;
   if( status == CHAR_ACK )
      rc = 0;
   else if( status == CHAR_CR )
   {
      if( nc < (buf_size-1))      // append CR
         buf[nc++] = CHAR_CR;
      rc = -EAGAIN;               // signal caller another line is ready.
   }
   else 
      rc = -EIO;

   /* terminate the buffer */
   buf[nc] = CHAR_NUL;
   *r_nc = nc;

   /* clear read_status */
   dpr_clr_read_ready(s);

   return rc;
}

/*-------------------------------------------------------------------
** mem_flush() - write a char to the pmac using the DPR interface 
**-------------------------------------------------------------------
*/
static int mem_flush( 
   struct stuff_t *s        // driver data
)
{
   int done;
   unsigned long timeout = 0;

   if( !s->dpr_enable )
      return -EPERM;

   writeb( CHAR_CTRL_X, s->reg_cntl_char);

   while(  timeout++ < IO_TIMEOUT_CNT )
   {
      if( dpr_read_ready( s ))
      {
         done = (readw(s->reg_read_status) & 0x00ff) == CHAR_CTRL_X;
         dpr_clr_read_ready(s);
         if( done )
            return 0;
      }
      else
         udelay( IO_UDELAYS );
   }
   return -EIO;
}


/*-------------------------------------------------------------------------
** ms2jiffies() - returns number of jiffies to insure 
**    ms milliseconds has elapsed. Minimum is 1.
**-------------------------------------------------------------------------
*/
static int ms2jiffies( long ms )
{
   long j;
   j = (ms * HZ) / 1000;     // calculate number of jiffies
   if( j < 1 ) j = 1;        // minimum is 1.
   return j;
}

/*-------------------------------------------------------------------
**  msleep() - ms based sleep - resolution based on kernel HZ value.
**-------------------------------------------------------------------
*/
static void msleep( int ms )
{
   long jiffies;
   current->state = TASK_INTERRUPTIBLE;
   jiffies = ms2jiffies(ms);
   schedule_timeout(jiffies);
}


