/*****************************************************************
** pmac.h - header file for pmac device driver
******************************************************************
*/

#ifndef _PMAC_H
#define _PMAC_H

// ioctl() cmd defines
#define PMAC_IOCTL_MSLEEP                1 // sleep function to test semaphore code
#define PMAC_IOCTL_STRING_REVERSE        2 // reverse the order of a user string  
#define PMAC_IOCTL_REG_CMD               3 // perform command IO using PCI register interface
#define PMAC_IOCTL_MEM_PUTS              4 // write a command line to DPR interface.
#define PMAC_IOCTL_MEM_GETS              5 // Reads a respones from the DPR interface.
#define PMAC_IOCTL_MEM_CMD               6 // Does a MEM_GETS/MEM_PUTS
#define PMAC_IOCTL_MEM_FLUSH             7 // Does a mem flush
#define PMAC_IOCTL_MEM_GETBLK            8 // copy a block of MEM to the caller
#define PMAC_IOCTL_MEM_PUTBLK            9 // copy a block of user data to MEM 
#define PMAC_IOCTL_MEM_GET_MOTOR_DATA   10 // copy DPRAM motor data to caller

#define PMAC_IOCTL_T1            101 // 
#define PMAC_IOCTL_T2            102 // 

#define PMAC_DPR_MOTOR0       0x0074 //  PC DPR offsets for Motor0 to Motor7 status block
#define PMAC_DPR_MOTOR1       0x00A4 
#define PMAC_DPR_MOTOR2       0x00D4 
#define PMAC_DPR_MOTOR3       0x0104 
#define PMAC_DPR_MOTOR4       0x0134 
#define PMAC_DPR_MOTOR5       0x0164 
#define PMAC_DPR_MOTOR6       0x0194 
#define PMAC_DPR_MOTOR7       0x01C4 
#define PMAC_DPR_MOTOR_NBYTES     48 //  size in byte of motor status block
//
// Variable (read/write) buffer related defines - this driver's uses
// of the variable read/write buffer is 
//
#define PMAC_VB_CNTL_ADDR         0x1044 //  Variable Read/Write Buffer control area  (PCI Address)
#define PMAC_VB_DATA_ADDR         0x1140 //  VB memory for actual read/write buffer

#define PMAC_VB_Y          0x00000000 //  VB data type: Y
#define PMAC_VB_D          0x00100000 //  VB data type: L
#define PMAC_VB_X          0x00200000 //  VB data type: X
//
// this driver assumes we are using these default pmac values
//
#define I10_DEFAULT     3713707.0      // I10 default value - servo interrupt time
#define Ixx08_DEFAULT        96.0      // Ixx08 default value - position scale factor
#define Ixx09_DEFAULT        96.0      // Ixx09 default value - velocity loop scale factor

/*------------------------------------------------------------
** pmac_parms_t - a general structure used to pass arguments
**               & return values to the ioctl() call.
**------------------------------------------------------------
*/
struct pmac_parms_t
{
   int ret;             /* return code    */
   unsigned long l[4];  /* array of longs */
   void *        p[2];  /* pointers */
};

/*------------------------------------------------------------
** Stucture and definition used by the MotorDataReporting Buffer
** related code
**------------------------------------------------------------
*/

struct pmac_ss_t {            // Servo Status ( ?  1st 24 bit word  )
   ushort rapid_spd_sel : 1;  // B00 - RAPID MOVE SPEED SELECT (IXX90)
   ushort dac_sign_mag : 1;   // B01 - SIGN/MAGNITUDE SERVO (IXX96)
   ushort sw_capture : 1;     // B02 - SOFTWARE HOME CAPTURE (IXX97.0)
   ushort fe_capture : 1;     // B03 - CAPTURE ON FOLLOWING ERROR (IXX97.1)
   ushort handwheel_ena : 1;  // B04 - HANDWHEEL ENABLE FLAG (IXX06.0)
   ushort hw_mode : 1;        // B05 - HANDWHEEL MODE FLAG (IXX06.1)
   ushort phased_motor : 1;   // B06 - PHASED MOTOR ENABLE FLAG (IXX01.0)
   ushort yenc_phase : 1;     // B07 - Y PHASE ENCODER (IXX01.1)
   ushort user_servo : 1;     // B08 - USER WRITEN SERVO ENABLE (IXX59.0)
   ushort user_phase : 1;     // B09 - USER WRITEN PHASE ENABLE (IXX59.1)
   ushort home_search : 1;    // B10 - HOME IN PROGRESS FLAG
   ushort block_request : 1;  // B11 - BLOCK REQUEST FLAG
   ushort limit_stop : 1;     // B12 - Limit Stop Flag
   ushort desired_vel_0 : 1;  // B13 - Desired Velocity = 0
   ushort data_block_err : 1; // B14 - DATA BLOCK ERROR
   ushort dwelling : 1;       // B15 - Dwell Mode
   ushort integrator_ena : 1; // B16 - Ixx34
   ushort run_program : 1;    // B17 - MOVE TIMER ACTIVE
   ushort open_loop : 1;      // B18 - OPEN LOOP MODE
   ushort amp_enabled : 1;    // B19 - AMPLIFIER ENABLED FLAG
   ushort algo_ena : 1;       // B20 - EXTENDED ALGO ENABLE FLAG (I3300+50*N)
   ushort pos_limit : 1;      // B21 - POSITIVE POSITION LIMIT
   ushort neg_limit : 1;      // B22 - NEGATIVE POSITION LIMIT
   ushort activated : 1;      // B23 - Ixx00
   ushort pads : 8;           // B24..31 - Not Available
};

struct pmac_ms_t {              // Motor Status ( ?  1st 24 bit word  )
   ushort in_position : 1;      // B00 - IN POSITION
   ushort warn_ferr : 1;        // B01 - SOFT FOLLOWING ERROR
   ushort fatal_ferr : 1;       // B02 - FATAL FOLLOWING ERROR
   ushort amp_fault : 1;        // B03 - AMP FAULT ERROR
   ushort backlash_dir : 1;     // B04 - BACKLASH DIRECTION FLAG
   ushort amp_i2t_err : 1;      // B05 - I2T AMP FAULT
   ushort integral_ferr : 1;    // B06 - INTEGRATED FOLLOWING ERROR FAULT
   ushort triger_home_flg : 1;  // B07 - TRIGGER/HOME MOVE FLAG
   ushort phase_find_err : 1;   // B08 - PHASE FINDING ERROR FLAG
   ushort tbd09 : 1;            // B09 - TBD
   ushort home_complete : 1;    // B10 - HOME COMPLETE FLAG
   ushort stopped_on_limit : 1; // B11 - POS LIMIT STOP FLAG
   ushort: 1;                   // B12 - TBD
   ushort: 1;                   // B13 - TBD
   ushort: 1;                   // B14 - TBD
   ushort cs_assigned : 1;      // B15 - TBD
   ushort cs_def : 4;           // B16..19 - Coord. Sys. Axis Def
   ushort coord_sys : 4;        // B20..23 - MOTOR COORDINATE SYSTEM NUMBER (-1)
   ushort padm : 8;             // B24..31 - Not Available
};

struct pmac_MRB_raw_t       /* pmac's DPR Motor reporting buffer format in DPR */
{
   long      pepos[2 ];        // $6001D-1E - position following err
   long      pdac;             // $6001F    - previous DAC
   struct pmac_ss_t sstatus;   // $60020    - Servo Status
   struct pmac_ms_t mstatus;   // $60021    - Status
   long      bpos[2 ];         // $60022-23 - Postion Bias
   long      fvel;             // $60024    - Filtered/Average Velocity
   long      hwpos[2 ];        // $60025-26 - Master Position
   long      apos[2 ];         // $60027-28 - Net actual position
};

/*------------------------------------------------------------
** Stucture and definitions for DPR Variable Buffer code
**------------------------------------------------------------
*/

struct pmac_vb_cntl_t         // Variable (read/write) Control area
{
   ushort read_cntl;          // VB - Read Control word.
   ushort timestamp;          // VB - Time Stamp return by PMAC
   ushort read_buf_size;      // VB - Read Buffer Size (num_entries or longs)
   ushort read_buf_offset;    // VB - Read Buffer Offset (DSP units)
   ushort write_buf_size;     // VB - Write Buffer Size (num_entries or longs)
   ushort write_buf_offset;   // VB - Write Buffer Offset (DSP units)
};

//-----------------------------------------------------
// nice prototypes
//----------------------------------------------------

int pmac_msleep( int fd, int ms );
int pmac_string_reverse( int fd, char * buf );
int pmac_reg_cmd( int fd, char * cmd, char * rpy, int rpy_size );
int pmac_mem_puts( int fd, char * cmd );
int pmac_mem_gets( int fd, char * rpy, int rpy_size );
int pmac_mem_flush( int fd );
int pmac_mem_cmd( int fd, char * cmd, char * rpy, int rpy_size );
int pmac_mem_getblk( int fd, void * dest, int src, int nbytes );
int pmac_mem_putblk( int fd, int dest, void * src, int nbytes );
int pmac_mem_get_motor_data( int fd, void * dest, long * timestamp, int m0, int nmotors);
int pmac_t1( int fd );
int pmac_t2( int fd );

double dprLongFixed( long d[], double scale );
long dprShortFixed( long word );
ulong dprShortU( ulong word );

#endif  // _PMAC_H

